@@ -14,6 +14,8 @@
 
 "Transport layer"
 
+import sys
+
 from ncclient.transport.session import Session, SessionListener, NetconfBase
 from ncclient.transport.ssh import SSHSession
 from ncclient.transport.tls import TLSSession
@@ -29,6 +31,15 @@ __all__ = [
     'SessionCloseError',
     'NetconfBase',
     'SSHError',
-    'SSHUnknownHostError'
-
+    'SSHUnknownHostError',
 ]
+
+#
+# Windows does not support Unix domain sockets; assume all other platforms do
+#
+if sys.platform != 'win32':
+    try:
+        from ncclient.transport.unixSocket import UnixSocketSession
+        __all__.append('UnixSocketSession')
+    except Exception:
+        pass
