@@ -17,11 +17,11 @@ import base64
 import getpass
 import os
 import re
-import six
 import sys
 import socket
 import threading
 from binascii import hexlify
+from io import BytesIO as StringIO
 
 try:
     import selectors
@@ -33,24 +33,16 @@ from ncclient.logging_ import SessionLoggerAdapter
 
 import paramiko
 
-from ncclient.transport.errors import AuthenticationError, SessionCloseError, SSHError, SSHUnknownHostError, NetconfFramingError
+from ncclient.transport.errors import AuthenticationError, SSHError, SSHUnknownHostError
 from ncclient.transport.session import Session
-from ncclient.transport.session import NetconfBase
+from ncclient.transport.parser import DefaultXMLParser
 
 import logging
 logger = logging.getLogger("ncclient.transport.ssh")
 
 PORT_NETCONF_DEFAULT = 830
-PORT_SSH_DEFAULT = 22
 
 BUF_SIZE = 4096
-# v1.0: RFC 4742
-MSG_DELIM = "]]>]]>"
-MSG_DELIM_LEN = len(MSG_DELIM)
-# v1.1: RFC 6242
-END_DELIM = '\n##\n'
-
-TICK = 0.1
 
 #
 # Define delimiters for chunks and messages for netconf 1.1 chunk enoding.
@@ -85,19 +77,6 @@ def _colonify(fp):
     return finga
 
 
-if sys.version < '3':
-    def textify(buf):
-        return buf
-else:
-    def textify(buf):
-        return buf.decode('UTF-8')
-
-if sys.version < '3':
-    from six import StringIO
-else:
-    from io import BytesIO as StringIO
-
-
 class SSHSession(Session):
 
     "Implements a :rfc:`4742` NETCONF session over SSH."
@@ -113,147 +92,23 @@ class SSHSession(Session):
         self._channel_id = None
         self._channel_name = None
         self._buffer = StringIO()
-        # parsing-related, see _parse()
         self._device_handler = device_handler
-        self._parsing_state10 = 0
-        self._parsing_pos10 = 0
-        self._parsing_pos11 = 0
-        self._parsing_state11 = 0
-        self._expchunksize = 0
-        self._curchunksize = 0
-        self._inendpos = 0
-        self._size_num_list = []
         self._message_list = []
         self._closing = threading.Event()
+        self.parser = DefaultXMLParser(self)  # SAX or DOM parser
 
         self.logger = SessionLoggerAdapter(logger, {'session': self})
 
     def _dispatch_message(self, raw):
-        self.logger.info("Received:\n%s", raw)
+        # Provide basic response message
+        self.logger.info("Received message from host")
+        # Provide complete response from host at debug log level
+        self.logger.debug("Received:\n%s", raw)
         return super(SSHSession, self)._dispatch_message(raw)
 
     def _parse(self):
         "Messages ae delimited by MSG_DELIM. The buffer could have grown by a maximum of BUF_SIZE bytes everytime this method is called. Retains state across method calls and if a byte has been read it will not be considered again."
-        return self._parse10()
-
-    def _parse10(self):
-
-        """Messages are delimited by MSG_DELIM. The buffer could have grown by
-        a maximum of BUF_SIZE bytes everytime this method is called. Retains
-        state across method calls and if a chunk has been read it will not be
-        considered again."""
-
-        self.logger.debug("parsing netconf v1.0")
-        buf = self._buffer
-        buf.seek(self._parsing_pos10)
-        if MSG_DELIM in buf.read().decode('UTF-8'):
-            buf.seek(0)
-            msg, _, remaining = buf.read().decode('UTF-8').partition(MSG_DELIM)
-            msg = msg.strip()
-            if sys.version < '3':
-                self._dispatch_message(msg.encode())
-            else:
-                self._dispatch_message(msg)
-            # create new buffer which contains remaining of old buffer
-            self._buffer = StringIO()
-            self._buffer.write(remaining.encode())
-            self._parsing_pos10 = 0
-            if len(remaining) > 0:
-                # There could be another entire message in the
-                # buffer, so we should try to parse again.
-                self.logger.debug('Trying another round of parsing since there is still data')
-                self._parse10()
-        else:
-            # handle case that MSG_DELIM is split over two chunks
-            self._parsing_pos10 = buf.tell() - MSG_DELIM_LEN
-            if self._parsing_pos10 < 0:
-                self._parsing_pos10 = 0
-
-    def _parse11(self):
-
-        """Messages are split into chunks. Chunks and messages are delimited
-        by the regex #RE_NC11_DELIM defined earlier in this file. Each
-        time we get called here either a chunk delimiter or an
-        end-of-message delimiter should be found iff there is enough
-        data. If there is not enough data, we will wait for more. If a
-        delimiter is found in the wrong place, a #NetconfFramingError
-        will be raised."""
-
-        self.logger.debug("_parse11: starting")
-
-        # suck in whole string that we have (this is what we will work on in
-        # this function) and initialize a couple of useful values
-        self._buffer.seek(0, os.SEEK_SET)
-        data = self._buffer.getvalue()
-        data_len = len(data)
-        start = 0
-        self.logger.debug('_parse11: working with buffer of %d bytes', data_len)
-        while True and start < data_len:
-            # match to see if we found at least some kind of delimiter
-            self.logger.debug('_parse11: matching from %d bytes from start of buffer', start)
-            re_result = RE_NC11_DELIM.match(data[start:])
-            if not re_result:
-
-                # not found any kind of delimiter just break; this should only
-                # ever happen if we just have the first few characters of a
-                # message such that we don't yet have a full delimiter
-                self.logger.debug('_parse11: no delimiter found, buffer="%s"', data[start:].decode())
-                break
-
-            # save useful variables for reuse
-            re_start = re_result.start()
-            re_end = re_result.end()
-            self.logger.debug('_parse11: regular expression start=%d, end=%d', re_start, re_end)
-
-            # If the regex doesn't start at the beginning of the buffer,
-            # we're in trouble, so throw an error
-            if re_start != 0:
-                raise NetconfFramingError('_parse11: delimiter not at start of match buffer', data[start:])
-
-            if re_result.group(2):
-                # we've found the end of the message, need to form up
-                # whole message, save back remainder (if any) to buffer
-                # and dispatch the message
-                start += re_end
-                message = ''.join(self._message_list)
-                self._message_list = []
-                self.logger.debug('_parse11: found end of message delimiter')
-                self._dispatch_message(message)
-                break
-
-            elif re_result.group(1):
-                # we've found a chunk delimiter, and group(2) is the digit
-                # string that will tell us how many bytes past the end of
-                # where it was found that we need to have available to
-                # save the next chunk off
-                self.logger.debug('_parse11: found chunk delimiter')
-                digits = int(re_result.group(1))
-                self.logger.debug('_parse11: chunk size %d bytes', digits)
-                if (data_len-start) >= (re_end + digits):
-                    # we have enough data for the chunk
-                    fragment = textify(data[start+re_end:start+re_end+digits])
-                    self._message_list.append(fragment)
-                    start += re_end + digits
-                    self.logger.debug('_parse11: appending %d bytes', digits)
-                    self.logger.debug('_parse11: fragment = "%s"', fragment)
-                else:
-                    # we don't have enough bytes, just break out for now
-                    # after updating start pointer to start of new chunk
-                    start += re_start
-                    self.logger.debug('_parse11: not enough data for chunk yet')
-                    self.logger.debug('_parse11: setting start to %d', start)
-                    break
-
-        # Now out of the loop, need to see if we need to save back any content
-        if start > 0:
-            self.logger.debug(
-                '_parse11: saving back rest of message after %d bytes, original size %d',
-                start, data_len)
-            self._buffer = StringIO(data[start:])
-            if start < data_len:
-                self.logger.debug('_parse11: still have data, may have another full message!')
-                self._parse11()
-        self.logger.debug('_parse11: ending')
+        return self.parser._parse10()
 
     def load_known_hosts(self, filename=None):
 
@@ -307,7 +162,10 @@ class SSHSession(Session):
             look_for_keys       = True,
             ssh_config          = None,
             sock_fd             = None,
-            bind_addr           = None):
+            bind_addr           = None,
+            sock                = None,
+            keepalive           = None,
+            environment         = None):
 
         """Connect via SSH and initialize the NETCONF session. First attempts the publickey authentication method and then password authentication.
 
@@ -315,7 +173,7 @@ class SSHSession(Session):
 
         *host* is the hostname or IP address to connect to
 
-        *port* is by default 830 (PORT_NETCONF_DEFAULT), but some devices use the default SSH port of 22 (PORT_SSH_DEFAULT) so this may need to be specified
+        *port* is by default 830 (PORT_NETCONF_DEFAULT), but some devices use the default SSH port of 22 so this may need to be specified
 
         *timeout* is an optional timeout for socket connect
 
@@ -340,9 +198,15 @@ class SSHSession(Session):
         *sock_fd* is an already open socket which shall be used for this connection. Useful for NETCONF outbound ssh. Use host=None together with a valid sock_fd number
 
         *bind_addr* is a (local) source IP address to use, must be reachable from the remote device.
+        
+        *sock* is an already open Python socket to be used for this connection.
+
+        *keepalive* Turn on/off keepalive packets (default is off). If this is set, after interval seconds without sending any data over the connection, a "keepalive" packet will be sent (and ignored by the remote host). This can be useful to keep connections alive over a NAT.
+
+        *environment* a dictionary containing the name and respective values to set
         """
-        if not (host or sock_fd):
-            raise SSHError("Missing host or socket fd")
+        if not (host or sock_fd or sock):
+            raise SSHError("Missing host, socket or socket fd")
 
         self._host = host
 
@@ -354,11 +218,11 @@ class SSHSession(Session):
             config = paramiko.SSHConfig()
             with open(os.path.expanduser(ssh_config)) as ssh_config_file_obj:
                 config.parse(ssh_config_file_obj)
-            
+
             # Save default Paramiko SSH port so it can be reverted
             paramiko_default_ssh_port = paramiko.config.SSH_PORT
-            
-            # Change the default SSH port to the port specified by the user so expand_variables 
+
+            # Change the default SSH port to the port specified by the user so expand_variables
             # replaces %p with the passed in port rather than 22 (the defauld paramiko.config.SSH_PORT)
 
             paramiko.config.SSH_PORT = port
@@ -374,23 +238,26 @@ class SSHSession(Session):
                 username = config.get("user")
             if key_filename is None:
                 key_filename = config.get("identityfile")
-            if hostkey_verify:
-                userknownhostsfile = config.get("userknownhostsfile")
-                if userknownhostsfile:
-                    self.load_known_hosts(os.path.expanduser(userknownhostsfile))
             if timeout is None:
                 timeout = config.get("connecttimeout")
                 if timeout:
                     timeout = int(timeout)
 
+        if hostkey_verify:
+            userknownhostsfile = config.get("userknownhostsfile")
+            if userknownhostsfile:
+                self.load_known_hosts(os.path.expanduser(userknownhostsfile))
+            else:
+                self.load_known_hosts()
+
         if username is None:
             username = getpass.getuser()
 
-        if sock_fd is None:
+        if sock_fd is None and sock is None:
             proxycommand = config.get("proxycommand")
             if proxycommand:
                 self.logger.debug("Configuring Proxy. %s", proxycommand)
-                if not isinstance(proxycommand, six.string_types):
+                if not isinstance(proxycommand, str):
                   proxycommand = [os.path.expanduser(elem) for elem in proxycommand]
                 else:
                   proxycommand = os.path.expanduser(proxycommand)
@@ -413,12 +280,8 @@ class SSHSession(Session):
                     break
                 else:
                     raise SSHError("Could not open socket to %s:%s" % (host, port))
-        else:
-            if sys.version_info[0] < 3:
-                s = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
-                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, _sock=s)
-            else:
-                sock = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
+        elif sock is None:
+            sock = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
             sock.settimeout(timeout)
 
         self._transport = paramiko.Transport(sock)
@@ -442,13 +305,11 @@ class SSHSession(Session):
         elif self._host_keys:
             # Else set preferred host keys to those we possess for the host
             # (avoids situation where known_hosts contains a valid key for the host, but that key type is not selected during negotiation)
-            if port == PORT_SSH_DEFAULT:
-                known_hosts_lookup = host
-            else:
-                known_hosts_lookup = '[%s]:%s' % (host, port)
-            known_host_keys_for_this_host = self._host_keys.lookup(known_hosts_lookup)
+            known_host_keys_for_this_host = self._host_keys.lookup(host) or {}
+            host_port = '[%s]:%s' % (host, port)
+            known_host_keys_for_this_host.update(self._host_keys.lookup(host_port) or {})
             if known_host_keys_for_this_host:
-                self._transport._preferred_keys = [x.key.get_name() for x in known_host_keys_for_this_host._entries]
+                self._transport._preferred_keys = list(known_host_keys_for_this_host)
 
         # Connect
         try:
@@ -456,18 +317,15 @@ class SSHSession(Session):
         except paramiko.SSHException as e:
             raise SSHError('Negotiation failed: %s' % e)
 
-        server_key_obj = self._transport.get_remote_server_key()
-        fingerprint = _colonify(hexlify(server_key_obj.get_fingerprint()))
-
         if hostkey_verify:
+            server_key_obj = self._transport.get_remote_server_key()
+            fingerprint = _colonify(hexlify(server_key_obj.get_fingerprint()))
+
             is_known_host = False
 
             # For looking up entries for nonstandard (22) ssh ports in known_hosts
             # we enclose host in brackets and append port number
-            if port == PORT_SSH_DEFAULT:
-                known_hosts_lookup = host
-            else:
-                known_hosts_lookup = '[%s]:%s' % (host, port)
+            known_hosts_lookups = [host, '[%s]:%s' % (host, port)]
 
             if hostkey_b64:
                 # If hostkey specified, remote host /must/ use that hostkey
@@ -475,10 +333,10 @@ class SSHSession(Session):
                     is_known_host = True
             else:
                 # Check known_hosts
-                is_known_host = self._host_keys.check(known_hosts_lookup, server_key_obj)
+                is_known_host = any(self._host_keys.check(lookup, server_key_obj) for lookup in known_hosts_lookups)
 
             if not is_known_host and not unknown_host_cb(host, fingerprint):
-                raise SSHUnknownHostError(known_hosts_lookup, fingerprint)
+                raise SSHUnknownHostError(known_hosts_lookups[0], fingerprint)
 
         # Authenticating with our private key/identity
         if key_filename is None:
@@ -493,6 +351,9 @@ class SSHSession(Session):
         self._connected = True      # there was no error authenticating
         self._closing.clear()
 
+        if keepalive:
+            self._transport.set_keepalive(keepalive)
+
         # TODO: leopoul: Review, test, and if needed rewrite this part
         subsystem_names = self._device_handler.get_ssh_subsystem_names()
         for subname in subsystem_names:
@@ -500,6 +361,17 @@ class SSHSession(Session):
             self._channel_id = self._channel.get_id()
             channel_name = "%s-subsystem-%s" % (subname, str(self._channel_id))
             self._channel.set_name(channel_name)
+            if environment:
+                try:
+                    self._channel.update_environment(environment)
+                except paramiko.SSHException as e:
+                    self.logger.info("%s (environment update rejected)", e)
+                    handle_exception = self._device_handler.handle_connection_exceptions(self)
+                    # Ignore the exception, since we continue to try the different
+                    # subsystem names until we find one that can connect.
+                    # have to handle exception for each vendor here
+                    if not handle_exception:
+                        continue
             try:
                 self._channel.invoke_subsystem(subname)
             except paramiko.SSHException as e:
@@ -511,7 +383,10 @@ class SSHSession(Session):
                 if not handle_exception:
                     continue
             self._channel_name = self._channel.get_name()
-            self._post_connect()
+            self._post_connect(timeout)
+            # for further upcoming RPC responses, vendor can chose their
+            # choice of parser. Say DOM or SAX
+            self.parser = self._device_handler.get_xml_parser(self)
             return
         raise SSHError("Could not open connection, possibly due to unacceptable"
                        " SSH subsystem name.")
@@ -519,22 +394,41 @@ class SSHSession(Session):
     def _auth(self, username, password, key_filenames, allow_agent,
               look_for_keys):
         saved_exception = None
-
+        encoded_password = None if password is None else password.encode('utf-8')
+        
         for key_filename in key_filenames:
-            for cls in (paramiko.RSAKey, paramiko.DSSKey, paramiko.ECDSAKey):
-                try:
-                    key = cls.from_private_key_file(key_filename, password)
-                    self.logger.debug("Trying key %s from %s",
-                                      hexlify(key.get_fingerprint()),
-                                      key_filename)
-                    self._transport.auth_publickey(username, key)
-                    return
-                except Exception as e:
-                    saved_exception = e
-                    self.logger.debug(e)
+            try:
+                key = paramiko.PKey.from_path(key_filename, encoded_password)
+                self.logger.debug("Trying key %s from %s",
+                                  hexlify(key.get_fingerprint()),
+                                  key_filename)
+                self._transport.auth_publickey(username, key)
+                return
+            except Exception as e:
+                saved_exception = e
+                self.logger.debug(e)
 
         if allow_agent:
-            for key in paramiko.Agent().get_keys():
+            # resequence keys from agent using private key names
+            prepend_agent_keys=[]
+            append_agent_keys=list(paramiko.Agent().get_keys())
+
+            for key_filename in key_filenames:
+                pubkey_filename=key_filename.strip(".pub")+".pub"
+                try:
+                    file_key=paramiko.PublicBlob.from_file(pubkey_filename).key_blob
+                except (FileNotFoundError, ValueError):
+                    continue
+
+                for idx, agent_key in enumerate(append_agent_keys):
+                    if agent_key.asbytes() == file_key:
+                        self.logger.debug("Prioritising SSH agent key found in %s",key_filename )
+                        prepend_agent_keys.append(append_agent_keys.pop(idx))
+                        break
+
+            agent_keys=tuple(prepend_agent_keys+append_agent_keys)
+
+            for key in agent_keys:
                 try:
                     self.logger.debug("Trying SSH agent key %s",
                                       hexlify(key.get_fingerprint()))
@@ -549,26 +443,32 @@ class SSHSession(Session):
             rsa_key = os.path.expanduser("~/.ssh/id_rsa")
             dsa_key = os.path.expanduser("~/.ssh/id_dsa")
             ecdsa_key = os.path.expanduser("~/.ssh/id_ecdsa")
+            ed25519_key = os.path.expanduser("~/.ssh/id_ed25519")
             if os.path.isfile(rsa_key):
-                keyfiles.append((paramiko.RSAKey, rsa_key))
+                keyfiles.append(rsa_key)
             if os.path.isfile(dsa_key):
-                keyfiles.append((paramiko.DSSKey, dsa_key))
+                keyfiles.append(dsa_key)
             if os.path.isfile(ecdsa_key):
-                keyfiles.append((paramiko.ECDSAKey, ecdsa_key))
+                keyfiles.append(ecdsa_key)
+            if os.path.isfile(ed25519_key):
+                keyfiles.append(ed25519_key)
             # look in ~/ssh/ for windows users:
             rsa_key = os.path.expanduser("~/ssh/id_rsa")
             dsa_key = os.path.expanduser("~/ssh/id_dsa")
             ecdsa_key = os.path.expanduser("~/ssh/id_ecdsa")
+            ed25519_key = os.path.expanduser("~/ssh/id_ed25519")
             if os.path.isfile(rsa_key):
-                keyfiles.append((paramiko.RSAKey, rsa_key))
+                keyfiles.append(rsa_key)
             if os.path.isfile(dsa_key):
-                keyfiles.append((paramiko.DSSKey, dsa_key))
+                keyfiles.append(dsa_key)
             if os.path.isfile(ecdsa_key):
-                keyfiles.append((paramiko.ECDSAKey, ecdsa_key))
+                keyfiles.append(ecdsa_key)
+            if os.path.isfile(ed25519_key):
+                keyfiles.append(ed25519_key)
 
-        for cls, filename in keyfiles:
+        for filename in keyfiles:
             try:
-                key = cls.from_private_key_file(filename, password)
+                key = paramiko.PKey.from_path(filename, encoded_password)
                 self.logger.debug("Trying discovered key %s in %s",
                                   hexlify(key.get_fingerprint()), filename)
                 self._transport.auth_publickey(username, key)
@@ -591,54 +491,17 @@ class SSHSession(Session):
 
         raise AuthenticationError("No authentication methods available")
 
-    def run(self):
-        chan = self._channel
-        q = self._q
+    def _transport_read(self):
+        return self._channel.recv(BUF_SIZE)
 
-        def start_delim(data_len): return '\n#%s\n' % (data_len)
+    def _transport_write(self, data):
+        return self._channel.send(data)
 
-        try:
-            s = selectors.DefaultSelector()
-            s.register(chan, selectors.EVENT_READ)
-            self.logger.debug('selector type = %s', s.__class__.__name__)
-            while True:
+    def _transport_register(self, selector, event):
+        selector.register(self._channel, event)
 
-                # Will wakeup evey TICK seconds to check if something
-                # to send, more quickly if something to read (due to
-                # select returning chan in readable list).
-                events = s.select(timeout=TICK)
-                if events:
-                    data = chan.recv(BUF_SIZE)
-                    if data:
-                        self._buffer.seek(0, os.SEEK_END)
-                        self._buffer.write(data)
-                        if self._base == NetconfBase.BASE_11:
-                            self._parse11()
-                        else:
-                            self._parse10()
-                    elif self._closing.is_set():
-                        # End of session, expected
-                        break
-                    else:
-                        # End of session, unexpected
-                        raise SessionCloseError(self._buffer.getvalue())
-                if not q.empty() and chan.send_ready():
-                    self.logger.debug("Sending message")
-                    data = q.get()
-                    if self._base == NetconfBase.BASE_11:
-                        data = "%s%s%s" % (start_delim(len(data)), data, END_DELIM)
-                    else:
-                        data = "%s%s" % (data, MSG_DELIM)
-                    self.logger.info("Sending:\n%s", data)
-                    while data:
-                        n = chan.send(data)
-                        if n <= 0:
-                            raise SessionCloseError(self._buffer.getvalue(), data)
-                        data = data[n:]
-        except Exception as e:
-            self.logger.debug("Broke out of main loop, error=%r", e)
-            self._dispatch_error(e)
-            self.close()
+    def _send_ready(self):
+        return self._channel.send_ready()
 
     @property
     def host(self):
